from rb.complexity.complexity_index import ComplexityIndex
from rb.core.lang import Lang
from rb.core.text_element import TextElement
from rb.complexity.index_category import IndexCategory
from rb.complexity.measure_function import MeasureFunction
from rb.core.text_element_type import TextElementType
from rb.complexity.discourse.conn_type_enum import ConnTypeEnum
from typing import List, Callable, Set
from rb.utils.rblogger import Logger
from typing import Dict, List
logger = Logger.get_logger()


class Connector(ComplexityIndex):
    
    
    conn_dict: Dict[ConnTypeEnum, List[str]] = None
    
    def __init__(self, lang: Lang, conn_type: ConnTypeEnum,
            reduce_depth: int, reduce_function: MeasureFunction):

        ComplexityIndex.__init__(self, lang=lang, category=IndexCategory.MORPHOLOGY,
                                 abbr="Connectors", reduce_depth=reduce_depth,
                                 reduce_function=reduce_function)
        if Connector.conn_dict is None:
            Connector.parse_connector_list(lang)

        self.conn_type = conn_type


    @staticmethod
    def parse_connector_list(lang: Lang):

        if lang is Lang.RO:
            path_to_wordlist = 'resources/ro/wordlists/connectives_ro.txt'
        elif lang is Lang.EN:
            path_to_wordlist = 'resources/en/wordlists/connectives_en.txt'

        with open(path_to_wordlist, 'rt', encoding='utf-8') as f:
            Connector.conn_dict = {}
            for ct in ConnTypeEnum:
                Connector.conn_dict[ct] = []
            
            conn_type = None
            if lang is Lang.EN:
                for i, line in enumerate(f):
                    if line.find('coordinating_connectives') != -1:
                        conn_type = ConnTypeEnum.COORD
                    elif line.find('logical_connectors') != -1:
                        conn_type = ConnTypeEnum.LOGIC
                    elif line.find('semi_coordinators') != -1:
                        conn_type = ConnTypeEnum.SEMI
                    elif line.find('quasi_coordinators') != -1:
                        conn_type = ConnTypeEnum.QUASI
                    elif line.find('conjunctions') != -1:
                        conn_type = ConnTypeEnum.CONJ
                    elif line.find('disjunctions') != -1:
                        conn_type = ConnTypeEnum.DISJ
                    elif line.find('simple_subordinators') != -1:
                        conn_type = ConnTypeEnum.SIMPLE
                    elif line.find('complex_subordinators') != -1:
                        conn_type = ConnTypeEnum.COMPLEX
                    elif line.find('coordinating_conjuncts') != -1:
                        conn_type = ConnTypeEnum.CONJ
                    elif line.find('addition') != -1:
                        conn_type = ConnTypeEnum.ADDITION
                    elif line.find('contrasts') != -1:
                        conn_type = ConnTypeEnum.CONTRAST
                    elif line.find('sentence_linking') != -1:
                        conn_type = ConnTypeEnum.LINK
                    elif line.find('order') != -1:
                        conn_type = ConnTypeEnum.ORDER
                    elif line.find('reference') != -1:
                        conn_type = ConnTypeEnum.REFERENCE
                    elif line.find('reason_and_purpose') != -1:
                        conn_type = ConnTypeEnum.REASONANDPURPOSE
                    elif line.find('conditions') != -1:
                        conn_type = ConnTypeEnum.COND
                    elif line.find('concessions') != -1:
                        conn_type = ConnTypeEnum.CONCC
                    elif line.find('oppositions') != -1:
                        conn_type = ConnTypeEnum.OPPOSIT
                    elif line.find('temporal_connectors') != -1:
                        conn_type = ConnTypeEnum.TEMPORAL
                    elif line.find('conjuncts') != -1:
                        conn_type = ConnTypeEnum.CONJUNCTS
                    elif len(line.strip()) == 0:
                        continue
                    elif conn_type is not None:
                        Connector.conn_dict[conn_type].append(line.strip())
            elif lang is Lang.RO:
                for i, line in enumerate(f):
                    if line.find('conectori_coordonare') != -1:
                        conn_type = ConnTypeEnum.COORD
                    elif line.find('conectori_logici') != -1:
                        conn_type = ConnTypeEnum.LOGIC
                    elif line.find('semi_coordonatori') != -1:
                        conn_type = ConnTypeEnum.SEMI
                    elif line.find('cvasi_coordonatori') != -1:
                        conn_type = ConnTypeEnum.QUASI
                    elif line.find('conjuncții') != -1:
                        conn_type = ConnTypeEnum.CONJ
                    elif line.find('disjuncții') != -1:
                        conn_type = ConnTypeEnum.DISJ
                    elif line.find('subcoordonatori_simpli') != -1:
                        conn_type = ConnTypeEnum.SIMPLE
                    elif line.find('subcoordonatori_complecși') != -1:
                        conn_type = ConnTypeEnum.COMPLEX
                    elif line.find('conjuncții_coordonatoare') != -1:
                        conn_type = ConnTypeEnum.CONJ
                    elif line.find('adăugire') != -1:
                        conn_type = ConnTypeEnum.ADDITION
                    elif line.find('contraste') != -1:
                        conn_type = ConnTypeEnum.CONTRAST
                    elif line.find('propoziții_conectate') != -1:
                        conn_type = ConnTypeEnum.LINK
                    elif line.find('ordine') != -1:
                        conn_type = ConnTypeEnum.ORDER
                    elif line.find('referință') != -1:
                        conn_type = ConnTypeEnum.REFERENCE
                    elif line.find('motiv_scop') != -1:
                        conn_type = ConnTypeEnum.REASONANDPURPOSE
                    elif line.find('concesii') != -1:
                        conn_type = ConnTypeEnum.CONCC
                    elif line.find('opoziții') != -1:
                        conn_type = ConnTypeEnum.OPPOSIT
                    elif line.find('conectori temporali') != -1:
                        conn_type = ConnTypeEnum.TEMPORAL
                    elif line.find('conjuncții') != -1:
                        conn_type = ConnTypeEnum.CONJUNCTS
                    elif len(line.strip()) == 0:
                        continue
                    elif conn_type is not None:
                        Connector.conn_dict[conn_type].append(line.strip())

    def process(self, element: TextElement) -> float:
        return self.reduce_function(self.compute_above(element))

    def compute_below(self, element: TextElement) -> List[str]:
        if element.is_word() == True:
            res = []
            if element.text.lower() in Connector.conn_dict[self.conn_type]:
                res.append(element.text)
            return res
        elif element.depth <= self.reduce_depth:
            res = []
            for child in element.components:
                res += self.compute_below(child)
            return res

    def compute_above(self, element: TextElement) -> List[float]:
        if element.depth > self.reduce_depth:
            values = []
            for child in element.components:
                values += self.compute_above(child)
            element.indices[self] = self.reduce_function(values)
        elif element.depth == self.reduce_depth:
            element.indices[self] = len(self.compute_below(element))
            values = [len(self.compute_below(element))]
            element.indices[self] = self.reduce_function(values)
        else:
            logger.error('wrong reduce depth value.')
        return values

    def __repr__(self):
        return self.reduce_function_abbr + self.reduce_depth_abbr + self.abbr + "_" + self.conn_type.name.lower()



from rb.core.lang import Lang
from rb.utils.downloader import download_aoa

import csv
from copy import deepcopy

from typing import Dict


class AgeOfAcquisition():

    def __init__(self, lang: Lang):
        self.lang = lang
        download_aoa(lang)
        self.aoa_dict = self.read_aoa_file()


    """
    Function that populates a dictionary with data from the AoA file for a specific language.
    Returns a dictionary that has as the key the word and as the value another dict
        representing the type of AoA (example: Kuperman) and its value
    """
    def read_aoa_file(self) -> Dict[str, Dict[str, float]]:
        aoa_dict = {}
        headers = []
        with open('resources/{}/aoa/AoA.csv'.format(self.lang.value)) as csv_file:
            csv_reader = csv.reader(csv_file, delimiter=',')
            for index, row in enumerate(csv_reader):
                if index == 0: # delimitator -> ignore
                    continue
                if index == 1:
                    headers = deepcopy(row)
                    continue
                # else
                for i, elem in enumerate(row):
                    if i == 0:
                        aoa_dict[elem] = {}
                    else:
                        aoa_dict[row[0]][headers[i]] = float(elem) if elem else None

        return aoa_dict


    def get_bird_value(self, word: str) -> float:
        if word not in self.aoa_dict or "Bird" not in self.aoa_dict[word]:
            return None
        return self.aoa_dict[word]["Bird"]


    def get_bristol_value(self, word: str) -> float:
        if word not in self.aoa_dict or "Bristol" not in self.aoa_dict[word]:
            return None
        return self.aoa_dict[word]["Bristol"]


    def get_cortese_value(self, word: str) -> float:
        if word not in self.aoa_dict or "Cortese" not in self.aoa_dict[word]:
            return None
        return self.aoa_dict[word]["Cortese"]


    def get_kuperman_value(self, word: str) -> float:
        if word not in self.aoa_dict or "Kuperman" not in self.aoa_dict[word]:
            return None
        return self.aoa_dict[word]["Kuperman"]

    
    def get_shock_value(self, word: str) -> float:
        if word not in self.aoa_dict or "Shock" not in self.aoa_dict[word]:
            return None
        return self.aoa_dict[word]["Shock"]



from multiprocessing import Pool, cpu_count
from typing import Callable, Iterable, List, Tuple

from rb.complexity.index_category import IndexCategory
from rb.complexity.measure_function import MeasureFunction, average, standard_deviation
from rb.core.lang import Lang
from rb.core.document import Document
from rb.core.text_element import TextElement
from rb.core.text_element_type import TextElementType
from rb.utils.rblogger import Logger
from rb.similarity.lda import LDA
from rb.similarity.lsa import LSA
from rb.similarity.word2vec import Word2Vec
from joblib import Parallel, delayed
from rb.similarity.vector_model import VectorModel
from rb.cna.cna_graph import CnaGraph
from rb.utils.downloader import download_wordlist

logger = Logger.get_logger()



"""TODO check for indices which compute on empty set of values """
class ComplexityIndex():
    """General class for any complexity index
    
    Attributes
    ----------
    lang : Lang
        language where the index is applied
    category : IndexCategory
        type of the index e.g. SURFACE, SYNTAX
    abbr : str
        short string describing the index
    reduce_depth : TextElementType
        the depth (in the document) where the reduce_function is applied
        the index is applied recursivley on all the above element types from the document tree
    reduce_function : Callable[[List], float]
        a function to summarize the results of the index (average or standard deviation)

    Methods
    -------
    process(element: TextElement)
        computes the index, overwritten for each index
    
    __repr__()
        detailed string representation of the index, should overwritten by each index
    """

    IDENTITY = 0

    def __init__(self, lang: Lang, category: IndexCategory, abbr: str, reduce_depth: int, reduce_function: MeasureFunction):
        self.lang = lang
        self.category = category
        self.abbr = abbr
        self.reduce_function = reduce_function
        self.reduce_depth = reduce_depth
        if self.reduce_function is None:
            self.reduce_function_abbr =  '' 
        elif self.reduce_function is average:
            self.reduce_function_abbr = 'Avg'
        elif self.reduce_function is standard_deviation:
            self.reduce_function_abbr =  'StDev'
        self.reduce_depth_abbr = '' if self.reduce_depth is None else self.element_to_abr(
                self.element_type_from_depth(self.reduce_depth).name)

    def element_type_from_depth(self, depth) -> TextElementType:
        for el_type in TextElementType:
            if el_type.value == depth:
                return el_type

    def element_to_abr(self, s) -> str:
        return s[0].upper() + s[1:].lower()

    # overwritten by each index
    def process(self, element: TextElement) -> float:
        pass

    # overwritten by each index 
    def __repr__(self):
        return self.reduce_function_abbr + self.reduce_depth_abbr + self.abbr

def compute_index(index: ComplexityIndex, element: TextElement) -> float:
    return index.process(element)

# computed indices and saves for each TextElement in indices dictionary
def compute_indices(doc: Document, cna_graph: CnaGraph = None):
    logger.info('Starting computing all indices for {0} type element'.format(type(doc).__name__))
    download_wordlist(doc.lang)
    num_cores = cpu_count()
    Parallel(n_jobs=num_cores, prefer="threads")(delayed(compute_index)(index, doc) \
        for cat in IndexCategory for index in cat.create(doc.lang, cna_graph))
        


from typing import List

from rb.core.block import Block
from rb.core.lang import Lang
from rb.core.sentence import Sentence
from rb.core.text_element import TextElement
from rb.core.text_element_type import TextElementType
from rb.core.word import Word
from rb.similarity.vector_model import (CorporaEnum, VectorModel,
                                        VectorModelType)
from rb.utils.downloader import download_tags


class Document(TextElement):
    

    def __init__(self, lang: Lang, text: str,
                 depth: int = TextElementType.DOC.value,
                 container: TextElement = None):
        TextElement.__init__(self, lang=lang, text=text,
                             depth=depth, container=container)
        text = text.replace("\n\n", "\n")
        for block in text.split("\n"):
            self.components.append(Block(lang=lang, text=block.strip(),
                                         container=self))

    def get_words(self) -> List[Word]:
        return [word for block in self.components for sent in block.components for word in sent.components]

    def get_sentences(self) -> List[Sentence]:
        return [sent for block in self.components for sent in block.components]

    def get_blocks(self) -> List[Block]:
        return self.components

    def __str__(self):
        return self.text



from typing import List
from rb.core.lang import Lang
from rb.core.text_element_type import TextElementType
from rb.complexity.measure_function import MeasureFunction
from rb.cna.cna_graph import CnaGraph


def create(lang: Lang, cna_graph: CnaGraph) -> List["ComplexityIndex"]:
    from rb.complexity.surface.no_words import NoWord
    from rb.complexity.surface.no_unq_words import NoUniqueWordsIndex
    from rb.complexity.surface.no_commas import NoCommas
    from rb.complexity.surface.no_punctuations import NoPunctuations
    from rb.complexity.surface.no_sentences import NoSentences
    from rb.complexity.surface.wd_entropy import WdEntropy
    from rb.complexity.surface.ch_entropy import ChEntropy
    from rb.complexity.surface.ch_ngram_entropy import ChNgramEntropyEnum, ChNgramEntropy

    indices = []

    indices.append(NoWord(lang, TextElementType.SENT.value, MeasureFunction.AVG))
    indices.append(NoWord(lang, TextElementType.SENT.value, MeasureFunction.STDEV))

    indices.append(NoWord(lang, TextElementType.BLOCK.value, MeasureFunction.AVG))
    indices.append(NoWord(lang, TextElementType.BLOCK.value, MeasureFunction.STDEV))

    indices.append(NoUniqueWordsIndex(lang, TextElementType.SENT.value, MeasureFunction.AVG))
    indices.append(NoUniqueWordsIndex(lang, TextElementType.SENT.value, MeasureFunction.STDEV))

    indices.append(NoSentences(lang, TextElementType.BLOCK.value, MeasureFunction.AVG))
    indices.append(NoSentences(lang, TextElementType.BLOCK.value, MeasureFunction.STDEV))

    indices.append(NoPunctuations(lang, TextElementType.SENT.value, MeasureFunction.AVG))
    indices.append(NoPunctuations(lang, TextElementType.SENT.value, MeasureFunction.STDEV))

    indices.append(NoPunctuations(lang, TextElementType.BLOCK.value, MeasureFunction.AVG))
    indices.append(NoPunctuations(lang, TextElementType.BLOCK.value, MeasureFunction.STDEV))

    indices.append(NoCommas(lang, TextElementType.SENT.value, MeasureFunction.AVG))
    indices.append(NoCommas(lang, TextElementType.SENT.value, MeasureFunction.STDEV))

    indices.append(NoCommas(lang, TextElementType.BLOCK.value, MeasureFunction.AVG))
    indices.append(NoCommas(lang, TextElementType.BLOCK.value, MeasureFunction.STDEV))

    indices.append(WdEntropy(lang, TextElementType.SENT.value, MeasureFunction.AVG))
    indices.append(WdEntropy(lang, TextElementType.SENT.value, MeasureFunction.STDEV))

    indices.append(WdEntropy(lang, TextElementType.BLOCK.value, MeasureFunction.AVG))
    indices.append(WdEntropy(lang, TextElementType.BLOCK.value, MeasureFunction.STDEV))

    indices.append(WdEntropy(lang, TextElementType.DOC.value, MeasureFunction.AVG))
    indices.append(WdEntropy(lang, TextElementType.DOC.value, MeasureFunction.STDEV))

    indices.append(ChEntropy(lang, TextElementType.WORD.value, MeasureFunction.AVG))
    indices.append(ChEntropy(lang, TextElementType.WORD.value, MeasureFunction.STDEV))

    indices.append(ChNgramEntropy(lang, ChNgramEntropyEnum.TWO, TextElementType.WORD.value, MeasureFunction.AVG))
    indices.append(ChNgramEntropy(lang, ChNgramEntropyEnum.TWO, TextElementType.WORD.value, MeasureFunction.STDEV))

    # if lang is Lang.RO:
    #     #indices.append(NoCacophonies(Lang.RO, TextElementType.SENT.value, MeasureFunction.AVG))
    #     indices.append(NoCommonErrors(Lang.RO, TextElementType.SENT.value, MeasureFunction.AVG))
    
    return indices

    


from typing import List

from rb.core.lang import Lang
from rb.complexity.measure_function import MeasureFunction
from rb.core.text_element_type import TextElementType
from rb.complexity.measure_function import MeasureFunction
from rb.cna.cna_graph import CnaGraph

# create all indices
# dependencies need to be putted in function because otherwise circular dependencies happens
def create(lang: Lang, cna_graph: CnaGraph) -> List["ComplexityIndex"]:
    from rb.complexity.rhythm.syllab import Syllab

    indices = []
    indices.append(Syllab(lang, TextElementType.WORD.value, MeasureFunction.AVG))
    return indices


from enum import Enum, auto
from rb.core.pos_features.pos_feature import POSFeature
from rb.core.pos_features.ro_pos_features.ro_features_name import RoFeaturesName
from rb.core.lang import Lang
from rb.core.pos import POS
from typing import List
import re


class RoPolarityEnum(Enum):
    NEG = auto()
    POS = auto()

class RoPOSFeaturePolarity(POSFeature):


    _INSTANCE = None

    def __init__(self):
        POSFeature.__init__(self, lang=Lang.RO, name=RoFeaturesName.POL,
                            values=None, pos_supported=None)
        self.values = set([ro_polarity for _, ro_polarity in RoPolarityEnum.__members__.items()])
        self.pos_supported = set([POS.ADV, POS.PART, POS.CCONJ, POS.SCONJ, POS.X, POS.ADP])

    @classmethod
    def get_instance(cls) -> "RoPOSFeaturePolarity":
        if cls._INSTANCE is None:
            cls._INSTANCE = RoPOSFeaturePolarity()
        return cls._INSTANCE

    def get_values(self, tag: str) -> List[RoPolarityEnum]:
        pattern = '(Polarity=[a-zA-Z0-9,]*){1}'
        matches = re.findall(pattern, tag)
        if len(matches) == 0: return []
        nrs = matches[0].split('=')[1].split(',')
        values = []
        for nr in nrs:
            if nr == 'Neg':
                values.append(RoPolarityEnum.NEG)
            elif nr == 'Pos':
                values.append(RoPolarityEnum.POS)
        return values


import setuptools
from setuptools.command.develop import develop
from setuptools.command.install import install
from subprocess import check_call

def do_post_install_tasks():
    # download spacy thing
   
    check_call("pip3 install https://github.com/explosion/spacy-models/releases/download/xx_ent_wiki_sm-2.1.0/xx_ent_wiki_sm-2.1.0.tar.gz".split())
    # download nltk stuff
    from os import getenv, path

    import nltk

    from rb.core.lang import Lang
    from rb.utils.downloader import download_wordnet

    nltk.download('punkt')
    nltk.download('wordnet')
    nltk.download('omw')

    nltk_path = getenv('NLTK_DATA')
    if nltk_path is None:
        nltk_path = path.expanduser('~/nltk_data/')
    download_wordnet(Lang.RO, nltk_path + "corpora/omw")


class PostDevelopmentCommand(develop):
    """Post-instalation for development? mode."""
    def run(self):
        do_post_install_tasks()
        # run
        develop.run(self)


class PostInstallCommand(install):
    """Post-instalation for install mode."""
    def run(self):
        do_post_install_tasks()
        install.run(self)

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name='rbpy-rb',
    version='0.6.17',
    author='Woodcarver',
    author_email='batpepastrama@gmail.com',
    description='ReaderBench library written in python',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://git.readerbench.com/ReaderBench/Readerbench-python',
    packages=setuptools.find_packages(),
    include_package_data=True,
    dependency_links=['https://github.com/explosion/spacy-models/releases/download/en_core_web_sm-2.1.0/en_core_web_sm-2.1.0.tar.gz'],
    install_requires=[
        'spacy==2.1.3',
        'pymorphy2-dicts==2.4.393442.3710985',
        'pymorphy2',
        'nltk==3.4.5',
        'gensim==3.8.1',
        'tensorflow==1.14',
        'sklearn',
        'networkx',
        'joblib',
        'blis<0.3.0',
        'boto',
        'boto3',
        'botocore',
        'certifi',
        'chardet',
        'Click',
        'cymem',
        'DAWG-Python',
        'decorator',
        'docopt',
        'docutils',
        'Flask',
        'idna',
        'itsdangerous',
        'Jinja2',
        'jmespath',
        'jsonschema<3.0.0',
        'MarkupSafe',
        'murmurhash',
        'neuralcoref',
        'plac<1.0.0',
        'preshed<2.1.0',
        'Pyphen',
        'python-dateutil<2.8.1',
        'requests',
        's3transfer',
        'scipy',
        'six',
        'smart-open',
        'srsly',
        'thinc<7.1.0',
        'tqdm',
        'urllib3',
        'wasabi',
        'Werkzeug',
        'wget',
        'pyLDAvis',
        'keras',
        'unidecode'
      ],
    cmdclass={
        'develop': PostDevelopmentCommand,
        'install': PostInstallCommand,
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent"
    ],
)



from rb.comprehension.utils.graph.cm_edge_type import CmEdgeType
from rb.comprehension.utils.graph.cm_node_do import CmNodeDO


class CmEdgeDO:

    def __init__(self, node1: CmNodeDO, node2: CmNodeDO, edge_type: CmEdgeType,
            score: float) -> None:
        self.node1 = node1
        self.node2 = node2
        self.edge_type = edge_type
        self.score = score
        self.active = True


    def get_node_1(self) -> CmNodeDO:
        return self.node1


    def get_node_2(self) -> CmNodeDO:
        return self.node2


    def get_edge_type(self) -> CmEdgeType:
        return self.edge_type


    def get_score(self) -> float:
        return self.score


    def is_active(self) -> bool:
        return self.active


    def activate(self) -> None:
        self.active = True


    def deactivate(self) -> None:
        self.active = False


    def get_opposite_node(self, node: CmNodeDO) -> CmNodeDO:
        if self.node1 == node:
            return self.node2
        elif self.node2 == node:
            return self.node1
        return None


    def __repr__(self):
        return self.node1.get_word().lemma + " - " + self.node2.get_word().lemma \
                + " ("+ str(self.edge_type) + ")" + ": " + str(self.score) + \
                " " + str(self.active)


    def __str__(self):
        return self.node1.get_word().lemma + " - " + self.node2.get_word().lemma \
                + " ("+ str(self.edge_type) + ")" + ": " + str(self.score) + \
                " " + str(self.active)


    def __eq__(self, other):
        if isinstance(other, CmEdgeDO):
            return ((self.node1 == other.node1 and self.node2 == other.node2) \
                    or (self.node2 == other.node1 and self.node1 == other.node2)) \
                    and self.edge_type == other.edge_type and self.score == other.score
        return NotImplemented


    def __hash__(self):
        return hash(tuple(self.node1, self.node2, self.score))




from enum import Enum, auto
from rb.core.pos_features.pos_feature import POSFeature
from rb.core.pos_features.ro_pos_features.ro_features_name import RoFeaturesName
from rb.core.lang import Lang
from rb.core.pos import POS
from typing import List
import re


class RoDefiniteEnum(Enum):
    DEF = auto()
    IND = auto()


class RoPOSFeatureDefinite(POSFeature):


    _INSTANCE = None

    def __init__(self):
        POSFeature.__init__(self, lang=Lang.RO, name=RoFeaturesName.DEF,
                            values=None, pos_supported=None)
        self.values = set([ro_definite for _, ro_definite in RoDefiniteEnum.__members__.items()])
        self.pos_supported = set([POS.ADJ, POS.DET, POS.NOUN, POS.NUM, POS.PROPN])
    
    @classmethod
    def get_instance(cls) -> "RoPOSFeatureDefinite":
        if cls._INSTANCE is None:
            cls._INSTANCE = RoPOSFeatureDefinite()
        return cls._INSTANCE

    def get_values(self, tag: str) -> List[RoDefiniteEnum]:
        pattern = '(Definite=[a-zA-Z,]*){1}'
        matches = re.findall(pattern, tag)
        if len(matches) == 0: return []
        definites = matches[0].split('=')[1].split(',')
        values = []
        for defi in definites:
            if defi == 'Def':
                values.append(RoDefiniteEnum.DEF)
            elif defi == 'Ind':
                values.append(RoDefiniteEnum.IND)
        return values


from typing import List, Callable, Dict
import csv
from rb.complexity.complexity_index import ComplexityIndex
from rb.core.lang import Lang
from rb.core.text_element import TextElement
from rb.complexity.index_category import IndexCategory
from rb.complexity.measure_function import MeasureFunction
from rb.core.text_element_type import TextElementType
from rb.complexity.word.valence_type import ValenceTypeEnum
from rb.utils.rblogger import Logger

logger = Logger.get_logger()


class Valence(ComplexityIndex):


    valence_dict: Dict[str, Dict[str, float]] = None

    def __init__(self, lang: Lang, valence_type: ValenceTypeEnum,
                 reduce_depth: int, reduce_function: MeasureFunction):

        ComplexityIndex.__init__(self, lang=lang, category=IndexCategory.WORD,
                                 abbr="Valence", reduce_depth=reduce_depth,
                                 reduce_function=reduce_function)
        if Valence.valence_dict is None:
            Valence.parse_valence_csv(lang)
        self.valence_type = valence_type

    @staticmethod
    def parse_valence_csv(lang: Lang):

        if lang is Lang.EN:
            path_to_valence = 'resources/en/wordlists/valences_en_with_liwc.csv'
        Valence.valence_dict = {}
        for vt in ValenceTypeEnum:
            Valence.valence_dict[vt] = {}

        with open(path_to_valence) as csv_file:
            csv_reader = csv.reader(csv_file, delimiter=';')
            for i, row in enumerate(csv_reader):
                if i < 2:
                    continue
                row[0] = row[0].strip()                    
                for j, vt in enumerate(ValenceTypeEnum):
                    if len(str(row[j])) > 0 and str(row[j + 1]) != '0':
                        Valence.valence_dict[vt][row[0]] = float(row[j + 1])

    def process(self, element: TextElement) -> float:
        return self.reduce_function(self.compute_above(element))

    def compute_below(self, element: TextElement) -> List[float]:
        if element.is_word() == True:
            if element.text in Valence.valence_dict[self.valence_type]:
                return [Valence.valence_dict[self.valence_type][element.text]]
            else:
                return []
        elif element.depth <= self.reduce_depth:
            res = []
            for child in element.components:
                res += self.compute_below(child)
            return res
    
    def compute_above(self, element: TextElement) -> List[float]:
        if element.depth > self.reduce_depth:
            values = []
            for child in element.components:
                values += self.compute_above(child)
            element.indices[self] = self.reduce_function(values)
        elif element.depth == self.reduce_depth:
            values = [len(self.compute_below(element))]
            element.indices[self] = self.reduce_function(values)
        else:
            logger.error('wrong reduce depth value.')
        return values
    
    def __repr__(self):
        return self.reduce_function_abbr + self.reduce_depth_abbr + self.abbr + "_" + self.valence_type.name.lower()


from rb.complexity.complexity_index import ComplexityIndex
from rb.core.lang import Lang
from rb.core.text_element import TextElement
from rb.complexity.index_category import IndexCategory
from rb.complexity.measure_function import MeasureFunction
from rb.core.text_element_type import TextElementType   
from typing import List, Callable
from rb.similarity.vector_model import VectorModel
from rb.cna.cna_graph import CnaGraph
from rb.cna.edge_type import EdgeType

from rb.utils.rblogger import Logger

logger = Logger.get_logger()


class TransCohesion(ComplexityIndex):

    def __init__(self, lang: Lang,
            reduce_depth: int = None, reduce_function: MeasureFunction = None,
            cna_graph: CnaGraph = None):
        ComplexityIndex.__init__(self, lang=lang, category=IndexCategory.COHESION,
                                 reduce_depth=reduce_depth, reduce_function=reduce_function,
                                 abbr="TransCohesion")
        self.cna_graph = cna_graph
        
    def process(self, element: TextElement) -> float:
        return self.compute(element)

    def compute(self, element: TextElement) -> float:

        blocks = element.get_blocks()
        if len(blocks) < 2:
            element.indices[self] = ComplexityIndex.IDENTITY
            return ComplexityIndex.IDENTITY            
        else:
            sim_values = []            
            for i, _ in enumerate(blocks[:-1]):
                current_block_sents = blocks[i].get_sentences()
                next_block_sents = blocks[i].get_sentences()
                if len(current_block_sents) == 0 or len(next_block_sents) == 0:
                    continue
                cur_sent = current_block_sents[-1]
                next_sent = next_block_sents[0]
                sim_edge = self.cna_graph.edges(node=(cur_sent, next_sent), edge_type=EdgeType.SEMANTIC,
                                                vector_model=None)
                v = sim_edge[0][2]
                sim_values.append(v)
            element.indices[self] = sum(sim_values) / len(sim_values)
            return element.indices[self]